#include "NebulaDevice.h"
#include "NebulaP2PTunnelAPIs.h"
#include "P2PTunnelServer.h"
#include "disposable_params.h"
#include <ctype.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <windows.h>

#define ENABLE_TUNNEL_SERVICE 0

#define HTTPS_SERVICE_PORT 443

#define NEBULA_BIND_TIMEOUT_MS 10000
#define MAX_CONNECTION 20

static int getNebulaProfileFromPath(const char *path, TUTKJsonObject **profile_obj) {
    int ret = 0;
    FILE *fp = NULL;
    char *profile = NULL;
    fp = fopen(path, "r");
    if (fp == NULL) {
        printf("Invalid path [%s] of profile\n", path);
        return -1;
    }

    fseek(fp, 0, SEEK_END);
    long fsize = ftell(fp);
    fseek(fp, 0, SEEK_SET);

    profile = malloc(fsize + 1);
    fread(profile, 1, fsize, fp);
    fclose(fp);
    profile[fsize] = '\0';

    ret = TUTK_Json_Obj_Create_From_String(profile, profile_obj);
    free(profile);
    if (ret != NEBULA_ER_NoERROR) {
        printf("Invalid profile\n");
        return -1;
    }
    return 0;
}

//########################################################
//# Nebula Device
//########################################################
static unsigned int gNebulaBindAbort = 0;
static char gPsk[MAX_NEBULA_PSK_LENGTH + 1] = {0};

static void WINAPI nebulaDeviceIdentityHandle(NebulaDeviceCtx *device, const char *identity, char *psk, unsigned int psk_size) {
    printf("[%s]%s\n", __func__, identity);
    if (psk_size > strlen(gPsk)) {
        strcpy(psk, gPsk);
    }
}

static int WINAPI nebulaDeviceCommandHandle(NebulaDeviceCtx *device, const char *identity, const char *func, const TUTKJsonObject *json_args, TUTKJsonObject **json_response) {
    printf("[%s] %s\n", __func__, func);
    // not implement in this sample
    return 400;
}

static int WINAPI nebulaDeviceSettingsChangeHandle(NebulaDeviceCtx *device, const char *settings) {
    printf("[%s]settings[%s]\n", __func__, settings);
    // not implement in this sample
    return 0;
}

static int WINAPI nebulaDeviceLoginStateHandle(NebulaDeviceCtx *device, NebulaDeviceLoginState state) {
    printf("[%s]login state[%d]\n", __func__, state);
    // not implement in this sample
    return 0;
}

static NebulaDeviceCtx *nebulaDeviceCreate(const char *udid, const char *secret_id, const char *pin_code, const char *psk, const char *profile_str) {
    int ret = 0;
    NebulaDeviceCtx *device_ctx = NULL;

    ret = Nebula_Initialize();
    printf("Nebula_Initialize() ret[%d]\n", ret);
    if (ret != NEBULA_ER_NoERROR) {
        return NULL;
    }

    ret = Nebula_Device_New(udid, secret_id, profile_str, nebulaDeviceCommandHandle, nebulaDeviceIdentityHandle, nebulaDeviceSettingsChangeHandle, &device_ctx);
    printf("Nebula_Device_New() ret[%d]\n", ret);
    if (ret != NEBULA_ER_NoERROR) {
        ret = Nebula_DeInitialize();
        printf("Nebula_DeInitialize() ret[%d]\n", ret);
        return NULL;
    }

    ret = Nebula_Device_Login(device_ctx, nebulaDeviceLoginStateHandle);
    printf("Nebula_Device_Login() ret[%d]\n", ret);
    if (ret != NEBULA_ER_NoERROR) {
        ret = Nebula_Device_Delete(device_ctx);
        printf("Nebula_Device_Delete() ret[%d]\n", ret);
        ret = Nebula_DeInitialize();
        printf("Nebula_DeInitialize() ret[%d]\n", ret);
        return NULL;
    }

    while (1) {
        ret = Nebula_Device_Bind(device_ctx, pin_code, gPsk, NEBULA_BIND_TIMEOUT_MS, &gNebulaBindAbort);
        printf("Nebula_Device_Bind() ret[%d]\n", ret);
        if (ret == NEBULA_ER_NoERROR) {
            break;
        } else if (ret != NEBULA_ER_TIMEOUT) {
            ret = Nebula_Device_Delete(device_ctx);
            printf("Nebula_Device_Delete() ret[%d]\n", ret);
            ret = Nebula_DeInitialize();
            printf("Nebula_DeInitialize() ret[%d]\n", ret);
            return NULL;
        }
    }

    return device_ctx;
}

//########################################################
//# IOTC Device
//########################################################

static int iotcDeviceCreateByNebula(NebulaDeviceCtx *device_ctx) {
    int ret = 0;

    while (1) {
        ret = IOTC_Device_Login_By_Nebula(device_ctx);
        printf("IOTC_Device_Login_By_Nebula() ret[%d]\n", ret);
        if (ret == IOTC_ER_NoERROR) {
            break;
        } else {
            Sleep(3000);
        }
    }
    return 0;
}

//########################################################
//# P2PTunnel Server
//########################################################
static bool gSidUsed[MAX_CONNECTION] = {0};

static void WINAPI tunnelSessionInfoExHandle(sP2PTunnelSessionInfoEx *sSessionInfo, const void *pArg) {
    int ret = 0;
    printf("[%s]\n", __func__);
    printf("  Connection Mode = %d, NAT type = %d\n", sSessionInfo->nMode, sSessionInfo->nNatType);
    printf("  P2PTunnel Version = %X, SID = %d\n", (unsigned int)sSessionInfo->nVersion, sSessionInfo->nSID);
    printf("  IP Address = %s:%d\n", sSessionInfo->szRemoteIP, sSessionInfo->nRemotePort);

    gSidUsed[sSessionInfo->nSID] = true;
    ret = P2PTunnel_SetBufSize(sSessionInfo->nSID, 5120000);
    if (ret != TUNNEL_ER_NoERROR) {
        printf("P2PTunnel_SetBufSize() SID[%d] error[%d]\n", sSessionInfo->nSID, ret);
    }

    return;
}

static void WINAPI tunnelStatusHandle(int nErrorCode, int nSID, void *pArg) {
    if (nErrorCode == TUNNEL_ER_DISCONNECTED) {
        printf("SID[%d] TUNNEL_ER_DISCONNECTED Log file here!\n", nSID);
        P2PTunnelServer_Disconnect(nSID);
        IOTC_Session_Close(nSID); // if AV module working on this session, AV will be disconnected.
        gSidUsed[nSID] = false;
    }
}

#if ENABLE_TUNNEL_SERVICE
static uint16_t WINAPI TunnelServiceVerifyCB(const char *cszServiceID, const void *pArg) {
    if (strcmp(cszServiceID, "https") == 0) {
        printf("[%s] service[%s] pass, service_port[%u]\n", __func__, cszServiceID, HTTPS_SERVICE_PORT);
        return HTTPS_SERVICE_PORT;
    }
    return 0;
}
#endif

static int tunnelServerManualListen(NebulaDeviceCtx *device_ctx) {

    int ret = 0;
    int nSID = IOTC_Device_Listen_By_Nebula(device_ctx, 5000);
    if (nSID > -1) {
        ret = P2PTunnelServer_Listen(nSID, 1);
        printf("P2PTunnelServer_Listen() ret = %d\n", ret);
    } else if (nSID == IOTC_ER_EXCEED_MAX_SESSION) {
        printf("IOTC_Listen() ret = IOTC_ER_EXCEED_MAX_SESSION\n");
    } else if (nSID == IOTC_ER_TIMEOUT) {
        printf("IOTC_Listen() ret = IOTC_ER_TIMEOUT\n");
    } else {
        printf("IOTC_Listen() error[%d]\n", nSID);
        return -1;
    }
    if (ret != TUNNEL_ER_NoERROR) {
        IOTC_Session_Close(nSID); // if AV module working on this session, AV will be disconnected.
    }
    return ret;
}

//########################################################
//# Main
//########################################################
int main(int argc, char *argv[]) {
    int ret = 0;
    char *udid = NULL;
    TUTKJsonObject *profile_obj = NULL;
    char *license_key = NULL;
    LogAttr log_attr = {0};
    char log_path[] = "./device_log.txt";
    NebulaDeviceCtx *device_ctx = NULL;
    char pin_code[MAX_PIN_CODE_LENGTH + 1] = {0};
    char secret_id[MAX_NEBULA_SECRETID_LENGTH + 1] = {0};

    if (argc != 4) {
        printf("%s <LICENSE_KEY> <UDID> <Profile>\n", argv[0]);
        return -1;
    }

    license_key = argv[1];
    udid = argv[2];

    if (strlen(udid) != MAX_UDID_LENGTH) {
        printf("invalid UDID\n");
        return -1;
    }

    //===== get profile for Nebula device =====
    if (getNebulaProfileFromPath(argv[3], &profile_obj) < 0) {
        return -1;
    }

    //===== get disposable params for Nebula connect =====
    ret = loadDisposableParams(pin_code, secret_id, gPsk);
    if (ret < 0) {
        printf("loadDisposableParams()=%d\n", ret);
        return -1;
    }

    //===== set license key for enable TUTK SDK =====
    ret = TUTK_SDK_Set_License_Key(license_key);
    printf("TUTK_SDK_Set_License_Key() ret[%d]\n", ret);
    if (ret < 0) {
        return -1;
    }

    //===== create Nebula device =====
    device_ctx = nebulaDeviceCreate(udid, secret_id, pin_code, gPsk, TUTK_Json_Obj_To_String(profile_obj));
    TUTK_Json_Obj_Release(profile_obj);
    if (device_ctx == NULL) {
        return -1;
    }

    //===== create IOTC device by Nebula =====
    IOTC_Set_Max_Session_Number(MAX_CONNECTION);
    ret = IOTC_Initialize2(0);
    if (ret != IOTC_ER_NoERROR) {
        printf("IOTC_Initialize2() ret[%d]\n", ret);
        Nebula_Device_Delete(device_ctx);
        Nebula_DeInitialize();
        return -1;
    }

    ret = iotcDeviceCreateByNebula(device_ctx);
    if (ret < 0) {
        printf("IOTC Login fail\n");
        IOTC_DeInitialize();
        Nebula_Device_Delete(device_ctx);
        Nebula_DeInitialize();
        return -1;
    }

    //==== create P2PTunnel server with manual mode =====
    log_attr.path = log_path;
    ret = P2PTunnel_Set_Log_Attr(log_attr);
    printf("P2PTunnel_Set_Log_Attr() ret[%d]\n", ret);

    ret = P2PTunnelSetConnectionOption(TUNNEL_CONNECT_MANUAL);
    if (ret != TUNNEL_ER_NoERROR) {
        printf("P2PTunnelSetConnectionOption() ret[%d]\n", ret);
        IOTC_DeInitialize();
        Nebula_Device_Delete(device_ctx);
        Nebula_DeInitialize();
        return -1;
    }

    ret = P2PTunnelServerInitialize2(MAX_CONNECTION, 0);
    if (ret != TUNNEL_ER_NoERROR) {
        printf("P2PTunnelServerInitialize2() ret[%d]\n", ret);
        IOTC_DeInitialize();
        Nebula_Device_Delete(device_ctx);
        Nebula_DeInitialize();
        return -1;
    }

    P2PTunnelServer_GetStatus(tunnelStatusHandle, NULL);

    ret = P2PTunnelServer_Start_By_Nebula(device_ctx, tunnelSessionInfoExHandle, NULL);
    if (ret != TUNNEL_ER_NoERROR) {
        printf("P2PTunnelServer_Start_By_Nebula() error[%d]!\n", ret);
        P2PTunnelServerDeInitialize();
        IOTC_DeInitialize();
        Nebula_Device_Delete(device_ctx);
        Nebula_DeInitialize();
        return -1;
    } else {
        printf("P2PTunnelServer_Start_By_Nebula() Success, I can connected by Internet.\n");
    }

#if ENABLE_TUNNEL_SERVICE
    ret = P2PTunnelServer_Register_Service_Verify(TunnelServiceVerifyCB, NULL);
    if (ret != TUNNEL_ER_NoERROR) {
        printf("P2PTunnelServer_Register_Service_Verify() error[%d]!\n", ret);
        return -1;
    }
#endif

    while (1) {
        ret = tunnelServerManualListen(device_ctx);
        for (int i = 0; i < MAX_CONNECTION; i++) {
            if (gSidUsed[i]) {
                int access_time = P2PTunnel_LastIOTime(i);
                if (access_time < 0) {
                    printf("P2PTunnel_LastIOTime() error[%d]\n", access_time);
                    gSidUsed[i] = false;
                } else {
                    printf("SID %d:%u, ", i, access_time);
                }
            }
        }
    }

    P2PTunnelServer_Stop();
    P2PTunnelServerDeInitialize();

    IOTC_DeInitialize();

    Nebula_Device_Delete(device_ctx);
    Nebula_DeInitialize();
    return 0;
}
